/* Syntaxpruefung fuer Regeln und Lexikon               */
/* E.Huckert  25.3.84                                   */
/*                                                      */
/* This program checks the syntax of rules and lexicons */
/* for HUpars                                           */
/*                                                      */
/* syncheck gehoert als Hilfsprogramm zum Parser HUpars   */
/* Standardannahmen:                                    */
/*    Hauptkategorien bestehen aus 4 Alpha - Zeichen    */
/*    Unterkategorien bestehen aus 3 Alpha - Zeichen    */
/*    UK- Werte von 1-8 sind zulaessig                  */
/*                                                      */
/* Das Lexikon ist aufsteigend sortiert                 */
/* Typischer Lexikoneintrag: PETER/EIGN NUM(1)GEN(1)    */
/*                                                      */
/* Die Regeln sind aufsteigend sortiert, wobei das      */
/* erste Element der linken Regelseite gewertet wird    */
/* Typische Regel:                                      */
/* ARTI NUM()GEN() + SUBS NUM()GEN() =                  */
/* NOGR NUM()GEN() ; (1,1,1)(2,2,2) .                   */
/*                                                      */
/* Es ist moeglich, dass syncheck nicht alle denkbaren  */
/* Fehler einer Grammatik oder eines Lexikons           */
/* entdeckt                                             */ 
/*                                                      */
/* Versionen                                            */
/* 18-07-2022  grob an modernen  ANSI C Parameterstil angepasst */
/* 08-09-2002  weitere Pruefungen eingebaut                     */
/*             Optionen -lexicon, -rules und -i eingebaut       */


#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#define CR   13
#define LF   10
#define DICT_SEP '='

/* Standard Laenge Hauptkategorie */
#define LAHK 4
/* Standard Laenge Unterkategorie */
#define LUNK 3

char frame[1000];
char hfrm[100];    // categorial fram without separators and values
char lentry;
char fe[100];
int  i, lahk, lunk, luk, lhk, ktei, level, anzuk;


// -----------------------------------------------------------------
/* einige globale Variablen initialisieren */
void iniv()
{
  lhk     = 0;
  luk     = 0;
  level   = 0;
  ktei    = 0;
  hfrm[0] = 0;
  anzuk   = 0;
  putchar(LF);
}   /* end iniv */

// -----------------------------------------------------------------
/* prueft, ob Zeichen in String           */
/* Liefert: die Position im String (ab 1) */
/*          sonst: 0                      */
int instr(char zeichen,
          char *str)
{
int n;
  n=0;
  while (*str != 0)
  {
    n++;
    if (zeichen == *str++) 
      return(n);
  }
  return(0);
}   /* end instr() */

// -----------------------------------------------------------------
/* Ein Zeichen an String haengen */
void append(char *str, 
            char  c)
{
  while (*str++ != 0);
  *--str = c;
  *++str = 0;
}   /* end append() */

// -----------------------------------------------------------------
/* Verkettung zweier Strings */
void konkat(char *ziel,
            char *quelle)
{
  /* akt. Laenge Ziel feststellen */
  while (*ziel++ !=0 );
  --ziel;
  while (*quelle != 0)
    *ziel++ = *quelle++;
  *ziel = 0;
}   /* end konkat */

// -----------------------------------------------------------------
/* String Compare ueber Laenge LAN */
/* Rueckgabe 0 bei Ungleichheit!   */
int scom(char *s1,
         char *s2,
         int   lan)
{
  while (lan-- > 0)
    if (*s1++ != *s2++) return(0);
  return(1);
}  /* end scom */

// -----------------------------------------------------------------
/* Fehlermeldung ausgeben und Programm verlassen */
void errexit(char *adr)
{
  printf("\n\n+++++ %s ++++++\n",adr);
  exit(-1);
}   /* end errexit */

// -----------------------------------------------------------------
/* Kategorialrahmen pruefen oder ablegen */
/* Liefert: 1 wenn OK, 0 wenn Fehler     */
int check_cat_frame(char *hfrm,
                    int  lhfrm)
{
char *adr;
  if (hfrm[0] == 0) 
    return(1);
  if (level != 0) 
    errexit("parentheses");
  /* Anz. der Unterkat. pruefen */
  if (anzuk > 8) 
    errexit("# subcategories > 8");
  adr = frame;
  /* Pruefen, ob Kat.rahmen schon existiert */
  while (*adr != 0)
  {
    if (scom(adr+1, hfrm, lhfrm) == 1)
    {
      /* HK identisch. Stimmt Rest auch ? */
      if (scom(adr+1, hfrm, *adr) == 0) 
        return(0);
      /* OK: Uebereinstimmung */
      return(1);
    }
    adr = adr + *adr + 1;
  }  /* while *adr != 0 */

  /* Kategorialrahmen ist neu */
  adr = hfrm;
  /* Kategorialrahmen mit Laenge eintragen */
  if (strlen(hfrm) > 0)
  {
    append(frame, (char)(strlen(hfrm)));
    konkat(frame, hfrm);
  }
  /* Alles OK */
  return(1);
}   /* end check_cat_frame() */

// -----------------------------------------------------------------
/* Sortierung im Lexikon pruefen */
// Returns: 0 if OK 
int wsort(char *word, char lentry)
{
  if (lentry > word[0]) 
    return(1);
  lentry  = word[0];
  word[0] = 0;
  return(0);
}   /* end wsort */
 
// -----------------------------------------------------------------
/* Sortierung pruefen */
// Returns: 0 if OK 
int sort(char *frame,
         char *last_frame)
{
  if (last_frame[0] > frame[0]) 
  {
    printf("\nsort error: last_entry=%s frame=%s\n", last_frame, frame);
    return(1);
  }
  return(0);
}   /* end sort */

// -----------------------------------------------------------------
/* Lexikon pruefen                                  */
/* Liefert: die Zahl der Eintraege (bis zum Fehler) */
int check_lexicon(FILE *fil)
{
int c;
int no_entries;
char word[128];
  //
  printf("\n\tcheck lexicon...\n");
  iniv() ;
  lentry     = 0;
  no_entries = 0;
  word[0]    = 0;

  while (1)
  {
    c = i = fgetc(fil);
    if (c == EOF) 
      break;
    if (c == CR) 
      continue;
    putchar(c);
    if (c == LF)
    {
      if (wsort(word, lentry) != 0) errexit("sort order");
      if (lhk != lahk)              errexit("length main category");
      if (level != 0)               errexit("number of parentheses");
      /* Kategorialrahmen pruefen */
      if (!check_cat_frame(hfrm, lahk))
         errexit("wrong categorial frame");
      iniv();
      continue;
    }
    if (c == DICT_SEP)
    {
      ktei = 1;
      no_entries++;
      continue;
    }
    if (ktei == 0)
    {
      append(word, (char)c);
      continue;
    }
    /* in Kategorienteil */
    if (ktei == 1)
    {
      /* In Hauptkategorie */
      if (c != ' ')
      {
        lhk++;
        append(hfrm, (char)c);
      }
      else
      {
        ktei = 2;
        luk  = 0;
      }
      continue;
    }
    /* in UK-Bereich (ktei = 2) */
    if (c == '(')
    {
      if (luk != lunk) errexit("length of subcategory");
      level++;
      anzuk++;
    }
    if (level == 0) luk++;
    if (c == ')')
    {
      level--;
      luk = 0;
    }
    if (instr((char)c, " (),12345678") == 0)
       append(hfrm, (char)c);
    if (instr((char)c, "09") > 0)
       errexit("invalid subcategory value");
  }   /* while (1) */
  //
  return no_entries;
}   /* end check_lexicon */

// -----------------------------------------------------------------
/* Regeln pruefen */
int check_rules(FILE *fil)
{
int rteil;
int count_plus, count_equal;
int count_rules;
char c;
char buf[128];
char last_hfrm[256];
  printf("\n\tcheck rules...\n");
  iniv();
  lentry  = 0;
  rteil   = 1;
  hfrm[0] = 0;
  last_hfrm[0] = 0;
  count_plus = 0;
  count_equal = 0;
  count_rules = 0;
  putchar(LF);

  while (1)
  {
    c = i = fgetc(fil);
    if (i == EOF) break;
    if (c == LF) continue;
    if (c == CR) continue;
    if (c == '$')
    {
      putchar(c);
      /* Kommentarzeile ignorieren */
      if (fgets(buf, sizeof(buf), fil) == NULL)
        break;
      printf("%s", buf);
      continue;
    } 
    putchar(c);
    if (c == '(') level++;
    if (c == ')') level--;
    if (c == '+')
    {
      count_plus++;
      rteil++;
      if (sort(hfrm, last_hfrm) != 0) 
        errexit("sort order");
      strcpy(last_hfrm, hfrm);
      if (! check_cat_frame(hfrm, lahk)) 
        errexit("categorial frame");
      iniv();
      continue;
    }
    if (c == '=')
    {
      rteil++;
      count_equal++;
      /* Einerregel? */
      if (rteil == 2)
      {
        /* das ist eine Einerregel */
        rteil = 3;
        if (sort(hfrm, last_hfrm) != 0) 
          errexit("sort order");
        strcpy(last_hfrm, hfrm);
      }
      if (! check_cat_frame(hfrm, lahk)) 
        errexit("categorial frame");
      iniv();
      continue;
    }
    if (c == ';')
    {
      rteil++;
      if (! check_cat_frame(hfrm, lahk)) 
        errexit("categorial frame");
      iniv();
      continue;
    }
    if (c == '.')
    {
      if (rteil != 4) 
        errexit("number of rule parts");
      if (level != 0) 
        errexit("parentheses");
      rteil = 1;
      if (count_plus > 1)
        errexit("more than 1 + character");
      if (count_equal > 1)
        errexit("more than 1 = character");
      iniv();
      count_plus = 0;
      count_equal = 0;
      count_rules++;
      continue; 
    }
    /* Zeichen im Kat.rahmen ablegen */
    if (rteil < 4)
    {
      if (instr(c," 12345678(),.;") == 0)
      {
        if (lhk >= 0) lhk++;
        if ((luk >= 0) && (lhk < 0)) luk++;
        append(hfrm, (char)c);
      }
      else
      {
        if (c == ' ')
        {
          /* Laenge Hauptkategorie pruefen */
          if (lhk > 0)
          {
            if (lhk != lahk) 
              errexit("length of main category");
            lhk = -1;
          }
        }   /* if c == ' ' */
        if (instr(c,"()") != 0)
        {
          if ((luk > 0) && (luk != lunk))
             errexit("length of subcategory");
          luk = 0;
        }
      }
      if (instr(c, "09") > 0)
        errexit("invalid subcategory value");
      if (c =='(') anzuk++;
      if (strlen(hfrm) > (unsigned)(80 + lahk))
        errexit("length of categorial frame");
    } 
  }   /* while (1) */
  //
  return count_rules;
}   /* end check_rules */

// -----------------------------------------------------------------
void usage()
{
  printf("Version 6 09-2002  -  HU\n");
  printf("usage: syncheck [-lexicon] [-rules] -i file\n");
}   // end usage()

// -----------------------------------------------------------------
int main(int argc, char *argv[])
{
char mode;
char filename[128];
FILE *inpfile;
  usage();
  
  /* Standardannahmen */
  lahk = LAHK;
  lunk = LUNK;
  
  /* uebergebene Argumente auswerten */
  mode = '\0';
  filename[0] = '\0';
  for (i=1; i < argc; i++)
  {
    if (strcmp(argv[i],"-h") == 0)
       sscanf(argv[i+1],"%d",&lahk);
    if (strcmp(argv[i],"-u") == 0)
       sscanf(argv[i+1],"%d",&lunk);
    if (strcmp(argv[i],"-lexicon") == 0)
       mode='L';
    if (strcmp(argv[i],"-rules") == 0)
       mode='R';
    if (strcmp(argv[i],"-i") == 0)
       strcpy(filename, argv[i + 1]);
  }  

  if (mode == '\0')
  {
    printf("ERROR: no mode specified\n");
    exit(3);
  }
  if (strlen(filename) == 0)
  {
    printf("ERROR: no filename specified\n");
    exit(1);
  }
  inpfile = fopen(filename, "r");
  if (inpfile == NULL)
  {
    printf("ERROR:cannot open the input file\n");
    exit(2);
  }

  frame[0] = 0;
  if (mode =='L')
  {
    int noe;
    noe = check_lexicon(inpfile);
    if (noe > 0)
      printf("\nOK: %d entries\n", noe);
  }
  if (mode == 'R')
  { 
    int noe;
    noe = check_rules(inpfile);
    if (noe > 0)
      printf("\nOK: %d rules\n", noe);
  }
  //
  fclose (inpfile);
  //
  return 0;
}   /* end main */
